"""The Kupo Python client."""

import os
from collections.abc import Callable
from collections.abc import Iterator
from enum import Enum
from functools import partial
from functools import wraps
from typing import Any
from typing import Self

import aiohttp as aio  # type: ignore
import requests
from dotenv import load_dotenv  # type: ignore
from pycardano import Address  # type: ignore
from pycardano import AssetName  # type: ignore
from pycardano import ScriptHash  # type: ignore
from pycardano import TransactionId  # type: ignore
from pydantic import BaseModel
from pydantic import ConfigDict
from pydantic import Field
from pydantic import RootModel
from pydantic import field_validator
from pydantic import model_validator

from kupo.models import Datum
from kupo.models import Health
from kupo.models import Match
from kupo.models import Metadata
from kupo.models import Pattern
from kupo.models import Point
from kupo.models import Script

load_dotenv()

KUPO_BASE_URL = os.environ.get("KUPO_BASE_URL", None)
KUPO_PORT = os.environ.get("KUPO_PORT", None)


class Order(Enum):
    """Order of results by time."""

    DESC = "most_recent_first"
    ASC = "least_recent_first"


class Limit(Enum):
    """Order of results by time."""

    SAFE = "within_safe_zone"
    UNSAFE = "unsafe_allow_beyond_safe_zone"


class QueryBase(BaseModel):
    """Query paramete base model, includes utility functions."""

    def params(self) -> dict[str, Any]:
        """Paramters for the query."""
        return self.model_dump(mode="json", exclude_none=True)


class BaseList(RootModel):
    """Base class for array responses."""

    def __getitem__(self, index: int) -> RootModel | str:
        """Get an item from the list."""
        return self.root[index]

    def __len__(self) -> int:
        """Get the length of the list."""
        return len(self.root)

    def __iter__(self) -> Iterator[RootModel | str]:
        """Iterate over the list."""
        return iter(self.root)


class MatchQuery(QueryBase):
    """Validator for the match query parameters."""

    model_config = ConfigDict(arbitrary_types_allowed=True)
    resolve_hashes: bool = Field(False, exclude=True)
    spent: bool = Field(False, exclude=True)
    unspent: bool = Field(False, exclude=True)
    order: Order = Order.DESC
    created_after: int | None = None
    spent_after: int | None = None
    created_before: int | None = None
    spent_before: int | None = None
    policy_id: str | ScriptHash | None = None
    asset_name: str | AssetName | None = None
    transaction_id: str | TransactionId | None = None
    output_index: int | None = None

    def flags(self) -> list[str]:
        """Get flags, which are empty parameters in the query string."""
        flags = []
        if self.spent:
            flags.append("spent")
        elif self.unspent:
            flags.append("unspent")

        if self.resolve_hashes:
            flags.append("resolve_hashes")

        return flags

    @field_validator(
        "spent_after",
        "spent_before",
        "created_after",
        "created_before",
        mode="before",
    )
    @classmethod
    def validate_int(cls, v: float | str) -> int | None:
        """Convert timestamps to integers."""
        return None if v is None else int(v)

    @field_validator(
        "policy_id",
        "asset_name",
        "transaction_id",
        mode="before",
    )
    @classmethod
    def validate_pycardano(
        cls,
        v: str | ScriptHash | AssetName | TransactionId | None,
    ) -> int | str | None:
        """Convert timestamps to integers."""
        if isinstance(v, ScriptHash | AssetName | TransactionId):
            v = v.payload.hex()
        return v

    @model_validator(mode="after")
    def validate_params(self) -> Self:
        """Run validation on match parameters to comply with API requirements."""
        if self.spent and self.unspent:
            raise ValueError("Cannot have both spent and unspent in parameters")

        if self.spent_after is not None and self.created_after is not None:
            raise ValueError(
                "Cannot have both spent_after and created_after in parameters",
            )

        if self.spent_before is not None and self.created_before is not None:
            raise ValueError(
                "Cannot have both spent_before and created_before in parameters",
            )

        if (
            self.spent_after is not None
            and self.spent_before is not None
            and self.spent_after > self.spent_before
        ):
            raise ValueError("spent_after must be less than spent_before")

        if (
            self.created_after is not None
            and self.created_before is not None
            and self.created_after > self.created_before
        ):
            raise ValueError("created_after must be less than created_before")

        if self.unspent and (
            self.spent_after is not None or self.spent_before is not None
        ):
            raise ValueError("Cannot have spent_after or spent_before with unspent")

        if (
            self.spent
            and self.spent_after is not None
            and self.spent_before is not None
        ):
            raise ValueError(
                "spent is redundant when spent_after and spent_before are set",
            )

        return self


class MatchResponse(BaseList):
    """The response model for endpoints that return matches."""

    root: list[Match]


class PatternResponse(BaseList):
    """The response model for endpoints that return patterns."""

    root: list[Pattern]


def rest_wrapper(
    func: Callable | None = None,
    param_model: type[BaseModel] | type[RootModel] | None = None,
    is_async: bool = False,
) -> Callable:
    """Autoparses input arguments to the given model."""
    if param_model is None:
        raise ValueError("Both param and response models must be supplied")

    if func is None:
        return partial(rest_wrapper, param_model=param_model, is_async=is_async)

    if is_async:

        @wraps(func)
        async def async_wrapped(*args, **kwargs):  # noqa
            if "model" not in kwargs or kwargs["model"] is None:
                parsed = param_model.model_validate(kwargs)
                kwargs["model"] = parsed
            return await func(*args, **kwargs)

        return async_wrapped

    @wraps(func)
    def wrapped(*args, **kwargs):  # noqa
        if "model" not in kwargs or kwargs["model"] is None:
            parsed = param_model.model_validate(kwargs)
            kwargs["model"] = parsed
        return func(*args, **kwargs)

    return wrapped


class KupoClient:
    """A Python client for the Kupo REST API."""

    def __init__(self, base_url: str | None = None, port: int | None = None) -> None:
        """The Python Kupo API client.

        Args:
            base_url: If not supplied, tries to read from the environment.
                Defaults to None.
            port: If not supplied, tries to read from the environment. Defaults to None.
        """
        if base_url is None:
            base_url = KUPO_BASE_URL

        if port is None and KUPO_PORT is not None:
            port = int(KUPO_PORT)

        if base_url is None:
            raise ValueError("base_url must be supplied")

        if port is None:
            raise ValueError("port must be supplied")

        self._base_url = base_url
        self._port = port

    @property
    def base_url(self) -> str:
        """The base URL for the client."""
        return f"{self._base_url}:{self._port}"

    async def _get_async(
        self,
        path: str,
        params: dict[str, Any] = {},
        flags: list[str] = [],
        timeout: int = 300,
    ) -> dict | list:
        path = self.base_url + path
        if len(flags) > 0:
            path += "?" + "&".join(flags)
        async with aio.ClientSession() as session:
            async with session.get(
                path,
                params=params,
                headers={"Accept": "application/json"},
                timeout=aio.ClientTimeout(total=timeout),
            ) as response:
                response.raise_for_status()
                return await response.json()

    def _get(
        self,
        path: str,
        params: dict[str, Any] = {},
        flags: list[str] = [],
        timeout: int = 300,
    ) -> dict | list:
        with requests.session() as session:
            path = self.base_url + path
            if len(flags) > 0:
                path += "?" + "&".join(flags)
            response = session.get(
                path,
                headers={"Accept": "application/json"},
                params=params,
                timeout=timeout,
            )
            response.raise_for_status()
            return response.json()

    async def _put_async(
        self,
        path: str,
        body: dict[str, Any] = {},
        timeout: int = 300,
    ) -> dict | list:
        async with aio.ClientSession() as session:
            path = self.base_url + path
            async with session.put(
                path,
                json=body,
                headers={"Accept": "application/json"},
                timeout=aio.ClientTimeout(total=timeout),
            ) as response:
                response.raise_for_status()
                return await response.json()

    def _put(
        self,
        path: str,
        body: dict[str, Any] = {},
        timeout: int = 300,
    ) -> dict | list:
        with requests.session() as session:
            path = self.base_url + path
            response = session.put(
                path,
                headers={"Accept": "application/json"},
                json=body,
                timeout=timeout,
            )
            response.raise_for_status()
            return response.json()

    @rest_wrapper(param_model=MatchQuery)
    def get_all_matches(
        self,
        resolve_hashes: bool = False,
        spent: bool = False,
        unspent: bool = False,
        order: Order = Order.DESC,
        created_after: int | str | None = None,
        spent_after: int | str | None = None,
        created_before: int | str | None = None,
        spent_before: int | str | None = None,
        policy_id: str | ScriptHash | None = None,
        asset_name: str | AssetName | None = None,
        transaction_id: str | TransactionId | None = None,
        output_index: int | None = None,
        timeout: int = 300,
        model: MatchQuery | None = None,
    ) -> MatchResponse:
        """Get all matches.

        Args:
            resolve_hashes: Resolve hashes to human readable strings. Defaults to False.
            spent: Only return spent transactions. Defaults to False.
            unspent: Only return unspect transactions. Defaults to False.
            order: Must be Orders.ASC or Orders.DESC. Defaults to Order.DESC.
            created_after: UTxO created after slot number. Defaults to None.
            spent_after: UTxO spent after slot number. Defaults to None.
            created_before: UTxO created before slot number. Defaults to None.
            spent_before: UTxO spent before slot number. Defaults to None.
            policy_id: UTxO contains token with policy id. Defaults to None.
            asset_name: UTxO contains token with policy name. Defaults to None.
            transaction_id: UTxOs with transaction id. Defaults to None.
            output_index: Index of output. Defaults to None.
            timeout: The timeout in seconds. Defaults to 300.
            model: The MatchQuery model. Will use if supplied. Defaults to None.

        Returns:
            _description_
        """
        if model is None:
            raise ValueError("model must be None")

        response = self._get(
            "/matches",
            params=model.params(),
            flags=model.flags(),
            timeout=timeout,
        )

        return MatchResponse.model_validate(response)

    @rest_wrapper(param_model=MatchQuery, is_async=True)
    async def get_all_matches_async(
        self,
        resolve_hashes: bool = False,
        spent: bool = False,
        unspent: bool = False,
        order: Order = Order.DESC,
        created_after: int | str | None = None,
        spent_after: int | str | None = None,
        created_before: int | str | None = None,
        spent_before: int | str | None = None,
        policy_id: str | ScriptHash | None = None,
        asset_name: str | AssetName | None = None,
        transaction_id: str | TransactionId | None = None,
        output_index: int | None = None,
        timeout: int = 300,
        model: MatchQuery | None = None,
    ) -> MatchResponse:
        """Get all matches.

        Args:
            resolve_hashes: Resolve hashes to human readable strings. Defaults to False.
            spent: Only return spent transactions. Defaults to False.
            unspent: Only return unspect transactions. Defaults to False.
            order: Must be Orders.ASC or Orders.DESC. Defaults to Order.DESC.
            created_after: UTxO created after slot number. Defaults to None.
            spent_after: UTxO spent after slot number. Defaults to None.
            created_before: UTxO created before slot number. Defaults to None.
            spent_before: UTxO spent before slot number. Defaults to None.
            policy_id: UTxO contains token with policy id. Defaults to None.
            asset_name: UTxO contains token with policy name. Defaults to None.
            transaction_id: UTxOs with transaction id. Defaults to None.
            output_index: Index of output. Defaults to None.
            timeout: The timeout in seconds. Defaults to 300.
            model: The MatchQuery model. Will use if supplied. Defaults to None.

        Returns:
            _description_
        """
        if model is None:
            raise ValueError("model must be None")

        response = await self._get_async(
            "/matches",
            params=model.params(),
            flags=model.flags(),
            timeout=timeout,
        )

        return MatchResponse.model_validate(response)

    @rest_wrapper(param_model=MatchQuery)
    def get_matches(
        self,
        pattern: str | Address,
        resolve_hashes: bool = False,
        spent: bool = False,
        unspent: bool = False,
        order: Order = Order.DESC,
        created_after: int | str | None = None,
        spent_after: int | str | None = None,
        created_before: int | str | None = None,
        spent_before: int | str | None = None,
        policy_id: str | ScriptHash | None = None,
        asset_name: str | AssetName | None = None,
        transaction_id: str | TransactionId | None = None,
        output_index: int | None = None,
        timeout: int = 300,
        model: MatchQuery | None = None,
    ) -> MatchResponse:
        """Get all matches.

        Args:
            pattern: The pattern to match.
            resolve_hashes: Resolve hashes to human readable strings. Defaults to False.
            spent: Only return spent transactions. Defaults to False.
            unspent: Only return unspect transactions. Defaults to False.
            order: Must be Orders.ASC or Orders.DESC. Defaults to Order.DESC.
            created_after: UTxO created after slot number. Defaults to None.
            spent_after: UTxO spent after slot number. Defaults to None.
            created_before: UTxO created before slot number. Defaults to None.
            spent_before: UTxO spent before slot number. Defaults to None.
            policy_id: UTxO contains token with policy id. Defaults to None.
            asset_name: UTxO contains token with policy name. Defaults to None.
            transaction_id: UTxOs with transaction id. Defaults to None.
            output_index: Index of output. Defaults to None.
            timeout: The timeout in seconds. Defaults to 300.
            model: The MatchQuery model. Will use if supplied. Defaults to None.

        Returns:
            The MatchResponse model.
        """
        if model is None:
            raise ValueError("model must be None")

        response = self._get(
            f"/matches/{pattern}",
            params=model.params(),
            flags=model.flags(),
            timeout=timeout,
        )

        return MatchResponse.model_validate(response)

    @rest_wrapper(param_model=MatchQuery, is_async=True)
    async def get_matches_async(
        self,
        pattern: str | Address,
        resolve_hashes: bool = False,
        spent: bool = False,
        unspent: bool = False,
        order: Order = Order.DESC,
        created_after: int | str | None = None,
        spent_after: int | str | None = None,
        created_before: int | str | None = None,
        spent_before: int | str | None = None,
        policy_id: str | ScriptHash | None = None,
        asset_name: str | AssetName | None = None,
        transaction_id: str | TransactionId | None = None,
        output_index: int | None = None,
        timeout: int = 300,
        model: MatchQuery | None = None,
    ) -> MatchResponse:
        """Get all matches.

        Args:
            pattern: The pattern to match.
            resolve_hashes: Resolve hashes to human readable strings. Defaults to False.
            spent: Only return spent transactions. Defaults to False.
            unspent: Only return unspect transactions. Defaults to False.
            order: Must be Orders.ASC or Orders.DESC. Defaults to Order.DESC.
            created_after: UTxO created after slot number. Defaults to None.
            spent_after: UTxO spent after slot number. Defaults to None.
            created_before: UTxO created before slot number. Defaults to None.
            spent_before: UTxO spent before slot number. Defaults to None.
            policy_id: UTxO contains token with policy id. Defaults to None.
            asset_name: UTxO contains token with policy name. Defaults to None.
            transaction_id: UTxOs with transaction id. Defaults to None.
            output_index: Index of output. Defaults to None.
            timeout: The timeout in seconds. Defaults to 300.
            model: The MatchQuery model. Will use if supplied. Defaults to None.

        Returns:
            The MatchResponse model.
        """
        if model is None:
            raise ValueError("model must be None")

        response = await self._get_async(
            f"/matches/{pattern}",
            params=model.params(),
            flags=model.flags(),
            timeout=timeout,
        )

        return MatchResponse.model_validate(response)

    def get_patterns(self, timeout: int = 300) -> PatternResponse:
        """Get all patterns used for matching.

        Args:
            timeout: The timeout in seconds. Defaults to 300.

        Returns:
            The list of matching patterns.
        """
        results = self._get("/patterns", timeout=timeout)

        return PatternResponse.model_validate(results)

    async def get_patterns_async(self, timeout: int = 300) -> PatternResponse:
        """Get all patterns used for matching.

        Args:
            timeout: The timeout in seconds. Defaults to 300.

        Returns:
            The list of matching patterns.
        """
        results = await self._get_async("/patterns", timeout=timeout)

        return PatternResponse.model_validate(results)

    def get_pattern(self, pattern: str, timeout: int = 300) -> PatternResponse:
        """Get all patterns that match a given pattern (if possible).

        Args:
            pattern: The pattern to match.
            timeout: The timeout in seconds. Defaults to 300.

        Returns:
            The list of matching patterns.
        """
        results = self._get(f"/patterns/{pattern}", timeout=timeout)

        return PatternResponse.model_validate(results)

    async def get_pattern_async(
        self,
        pattern: str,
        timeout: int = 300,
    ) -> PatternResponse:
        """Get all patterns that match a given pattern (if possible).

        Args:
            pattern: The pattern to match.
            timeout: The timeout in seconds. Defaults to 300.

        Returns:
            The list of matching patterns.
        """
        results = await self._get_async(f"/patterns/{pattern}", timeout=timeout)

        return PatternResponse.model_validate(results)

    def get_script_by_hash(self, script_hash: str, timeout: int = 300) -> Script:
        """Get a script by its hash.

        Args:
            script_hash: The hash of the script to fetch.
            timeout: The timeout in seconds. Defaults to 300.

        Returns:
            The script matching the hash.
        """
        results = self._get(f"/scripts/{script_hash}", timeout=timeout)

        return Script.model_validate(results)

    async def get_script_by_hash_async(
        self,
        script_hash: str,
        timeout: int = 300,
    ) -> Script:
        """Get a script by its hash.

        Args:
            script_hash: The hash of the script to fetch.
            timeout: The timeout in seconds. Defaults to 300.

        Returns:
            The script matching the hash.
        """
        results = await self._get_async(f"/scripts/{script_hash}", timeout=timeout)

        return Script.model_validate(results)

    def get_datum_by_hash(self, datum_hash: str, timeout: int = 300) -> Datum:
        """Get a datum by its hash.

        Args:
            datum_hash: The hash of the datum to fetch.
            timeout: The timeout in seconds. Defaults to 300.

        Returns:
            The datum matching the hash.
        """
        results = self._get(f"/datums/{datum_hash}", timeout=timeout)

        return Datum.model_validate(results)

    async def get_datum_by_hash_async(
        self,
        datum_hash: str,
        timeout: int = 300,
    ) -> Datum:
        """Get a datum by its hash.

        Args:
            datum_hash: The hash of the datum to fetch.
            timeout: The timeout in seconds. Defaults to 300.

        Returns:
            The datum matching the hash.
        """
        results = await self._get_async(f"/datums/{datum_hash}", timeout=timeout)

        return Datum.model_validate(results)

    def get_metadata_by_tx(self, tx_hash: str, timeout: int = 300) -> Metadata:
        """Get a metadata by transaction hash.

        Args:
            tx_hash: The hash of the metadata to fetch.
            timeout: The timeout in seconds. Defaults to 300.

        Returns:
            The metadata matching the hash.
        """
        results = self._get(f"/metadata/{tx_hash}", timeout=timeout)

        return Metadata.model_validate(results)

    async def get_metadata_by_tx_async(
        self,
        tx_hash: str,
        timeout: int = 300,
    ) -> Metadata:
        """Get a datum by its hash.

        Args:
            tx_hash: The hash of the metadata to fetch.
            timeout: The timeout in seconds. Defaults to 300.

        Returns:
            The metadata matching the hash.
        """
        results = await self._get_async(f"/metadata/{tx_hash}", timeout=timeout)

        return Metadata.model_validate(results)

    def add_pattern(
        self,
        pattern: str,
        rollback_to: int | dict | Point,
        limit: Limit = Limit.SAFE,
        timeout: int = 300,
    ) -> PatternResponse:
        """Add a pattern to the database.

        Args:
            pattern: The capture pattern.
            rollback_to: The point to rollback to. Must be either a slot number (int),
                a Point object, or a dictionary with keys `slot_no` (required) and
                `header_hash` (optional).
            limit: Use safe or unsafe boundaries as defined in the docs.
                Defaults to Limit.SAFE.
            timeout: Timeout for the request. Defaults to 300.

        Returns:
            A PatternResponse object.
        """
        if isinstance(rollback_to, Point):
            rollback_to = rollback_to.model_dump(mode="json")
        elif isinstance(rollback_to, int):
            rollback_to = {"slot_no": rollback_to}

        if "slot_no" not in rollback_to:
            raise ValueError("rollback_to must contain slot_no")

        results = self._put(
            f"/patterns/{pattern}",
            body={"rollback_to": rollback_to, "limit": limit.value},
            timeout=timeout,
        )

        return PatternResponse.model_validate(results)

    async def add_pattern_async(
        self,
        pattern: str,
        rollback_to: int | dict | Point,
        limit: Limit = Limit.SAFE,
        timeout: int = 300,
    ) -> PatternResponse:
        """Add a pattern to the database.

        Args:
            pattern: The capture pattern.
            rollback_to: The point to rollback to. Must be either a slot number (int),
                a Point object, or a dictionary with keys `slot_no` (required) and
                `header_hash` (optional).
            limit: Use safe or unsafe boundaries as defined in the docs.
                Defaults to Limit.SAFE.
            timeout: Timeout for the request. Defaults to 300.

        Returns:
            A PatternResponse object.
        """
        if isinstance(rollback_to, Point):
            rollback_to = rollback_to.model_dump(mode="json")
        elif isinstance(rollback_to, int):
            rollback_to = {"slot_no": rollback_to}

        if "slot_no" not in rollback_to:
            raise ValueError("rollback_to must contain slot_no")

        results = await self._put_async(
            f"/patterns/{pattern}",
            body={"rollback_to": rollback_to, "limit": limit.value},
            timeout=timeout,
        )

        return PatternResponse.model_validate(results)

    def bulk_add_pattern(
        self,
        patterns: list[str],
        rollback_to: int | dict | Point,
        limit: Limit = Limit.SAFE,
        timeout: int = 300,
    ) -> PatternResponse:
        """Add multiple patterns to the database.

        Args:
            patterns: The capture patterns.
            rollback_to: The point to rollback to. Must be either a slot number (int),
                a Point object, or a dictionary with keys `slot_no` (required) and
                `header_hash` (optional).
            limit: Use safe or unsafe boundaries as defined in the docs.
                Defaults to Limit.SAFE.
            timeout: Timeout for the request. Defaults to 300.

        Returns:
            A PatternResponse object.
        """
        if isinstance(rollback_to, Point):
            rollback_to = rollback_to.model_dump(mode="json")
        elif isinstance(rollback_to, int):
            rollback_to = {"slot_no": rollback_to}

        if "slot_no" not in rollback_to:
            raise ValueError("rollback_to must contain slot_no")

        results = self._put(
            "/patterns",
            body={
                "patterns": patterns,
                "rollback_to": rollback_to,
                "limit": limit.value,
            },
            timeout=timeout,
        )

        return PatternResponse.model_validate(results)

    async def bulk_add_pattern_async(
        self,
        patterns: list[str],
        rollback_to: int | dict | Point,
        limit: Limit = Limit.SAFE,
        timeout: int = 300,
    ) -> PatternResponse:
        """Add multiple patterns to the database.

        Args:
            patterns: The capture patterns.
            rollback_to: The point to rollback to. Must be either a slot number (int),
                a Point object, or a dictionary with keys `slot_no` (required) and
                `header_hash` (optional).
            limit: Use safe or unsafe boundaries as defined in the docs.
                Defaults to Limit.SAFE.
            timeout: Timeout for the request. Defaults to 300.

        Returns:
            A PatternResponse object.
        """
        if isinstance(rollback_to, Point):
            rollback_to = rollback_to.model_dump(mode="json")
        elif isinstance(rollback_to, int):
            rollback_to = {"slot_no": rollback_to}

        if "slot_no" not in rollback_to:
            raise ValueError("rollback_to must contain slot_no")

        results = await self._put_async(
            "/patterns",
            body={
                "patterns": patterns,
                "rollback_to": rollback_to,
                "limit": limit.value,
            },
            timeout=timeout,
        )

        return PatternResponse.model_validate(results)

    def health(self) -> Health:
        """Check the health of the API."""
        return Health.model_validate(self._get("/health"))

    async def health_async(self) -> Health:
        """Check the health of the API."""
        return Health.model_validate(await self._get_async("/health"))



import pytest


def test_get_all_matches(
    benchmark,
    kupo_client,
    spent,
    unspent,
    spent_after,
    created_after,
    spent_before,
    created_before,
):

    try:

        if spent_after is None and created_after is None:
            pytest.skip("Not having spent_after or created_after takes too long.")

        benchmark(
            kupo_client.get_all_matches,
            spent=spent,
            unspent=unspent,
            spent_after=spent_after,
            created_after=created_after,
            spent_before=spent_before,
            created_before=created_before,
        )
    except ValueError as e:
        if spent and unspent:
            pytest.xfail(f"Cannot have both spent and unspent in parameters: {e}")
        elif spent and (spent_after is not None or spent_before is not None):
            pytest.xfail(
                f"Cannot have spent and spent_after or spent_before in parameters: {e}"
            )
        elif unspent and (spent_after is not None or spent_before is not None):
            pytest.xfail(
                f"Cannot have unspent and spent_after or spent_before in parameters: {e}"
            )
        elif spent_after is not None and created_after is not None:
            pytest.xfail(
                f"Cannot have spent_after and created_after in parameters: {e}"
            )
        elif spent_before is not None and created_before is not None:
            pytest.xfail(
                f"Cannot have spent_before and created_before in parameters: {e}"
            )
        else:
            raise e


@pytest.mark.asyncio
async def test_get_all_matches_async(
    benchmark,
    kupo_client,
    spent,
    unspent,
    spent_after,
    created_after,
    spent_before,
    created_before,
):

    try:

        if spent_after is None and created_after is None:
            pytest.skip("Not having spent_after or created_after takes too long.")

        await kupo_client.get_all_matches_async(
            spent=spent,
            unspent=unspent,
            spent_after=spent_after,
            created_after=created_after,
            spent_before=spent_before,
            created_before=created_before,
        )
    except ValueError as e:
        if spent and unspent:
            pytest.xfail(f"Cannot have both spent and unspent in parameters: {e}")
        elif spent and (spent_after is not None or spent_before is not None):
            pytest.xfail(
                f"Cannot have spent and spent_after or spent_before in parameters: {e}"
            )
        elif unspent and (spent_after is not None or spent_before is not None):
            pytest.xfail(
                f"Cannot have unspent and spent_after or spent_before in parameters: {e}"
            )
        elif spent_after is not None and created_after is not None:
            pytest.xfail(
                f"Cannot have spent_after and created_after in parameters: {e}"
            )
        elif spent_before is not None and created_before is not None:
            pytest.xfail(
                f"Cannot have spent_before and created_before in parameters: {e}"
            )
        else:
            raise e


def test_get_matches(
    benchmark,
    kupo_client,
    pattern,
    spent,
    unspent,
    spent_after,
    created_after,
    spent_before,
    created_before,
):

    try:

        if spent_after is None and created_after is None:
            pytest.skip("Not having spent_after or created_after takes too long.")

        benchmark(
            kupo_client.get_matches,
            pattern=pattern,
            spent=spent,
            unspent=unspent,
            spent_after=spent_after,
            created_after=created_after,
            spent_before=spent_before,
            created_before=created_before,
        )
    except ValueError as e:
        if spent and unspent:
            pytest.xfail(f"Cannot have both spent and unspent in parameters: {e}")
        elif spent and (spent_after is not None or spent_before is not None):
            pytest.xfail(
                f"Cannot have spent and spent_after or spent_before in parameters: {e}"
            )
        elif unspent and (spent_after is not None or spent_before is not None):
            pytest.xfail(
                f"Cannot have unspent and spent_after or spent_before in parameters: {e}"
            )
        elif spent_after is not None and created_after is not None:
            pytest.xfail(
                f"Cannot have spent_after and created_after in parameters: {e}"
            )
        elif spent_before is not None and created_before is not None:
            pytest.xfail(
                f"Cannot have spent_before and created_before in parameters: {e}"
            )
        else:
            raise e


@pytest.mark.asyncio
async def test_get_matches_async(
    benchmark,
    kupo_client,
    pattern,
    spent,
    unspent,
    spent_after,
    created_after,
    spent_before,
    created_before,
):

    try:

        if spent_after is None and created_after is None:
            pytest.skip("Not having spent_after or created_after takes too long.")

        await kupo_client.get_matches_async(
            pattern=pattern,
            spent=spent,
            unspent=unspent,
            spent_after=spent_after,
            created_after=created_after,
            spent_before=spent_before,
            created_before=created_before,
        )
    except ValueError as e:
        if spent and unspent:
            pytest.xfail(f"Cannot have both spent and unspent in parameters: {e}")
        elif spent and (spent_after is not None or spent_before is not None):
            pytest.xfail(
                f"Cannot have spent and spent_after or spent_before in parameters: {e}"
            )
        elif unspent and (spent_after is not None or spent_before is not None):
            pytest.xfail(
                f"Cannot have unspent and spent_after or spent_before in parameters: {e}"
            )
        elif spent_after is not None and created_after is not None:
            pytest.xfail(
                f"Cannot have spent_after and created_after in parameters: {e}"
            )
        elif spent_before is not None and created_before is not None:
            pytest.xfail(
                f"Cannot have spent_before and created_before in parameters: {e}"
            )
        else:
            raise e



# generated by datamodel-codegen:
#   filename:  https://cardanosolutions.github.io/kupo/api/nightly.yaml
#   timestamp: 2025-01-07T02:57:18+00:00

from __future__ import annotations

from enum import Enum
from typing import Dict, List, Optional, Union

from pydantic import BaseModel, ConfigDict, Field, RootModel, conint, constr


class AddressShelley(RootModel[str]):
    root: str = Field(..., title="Shelley")


class AddressStake(RootModel[str]):
    root: str = Field(..., title="Stake")


class AddressBootstrap(RootModel[str]):
    root: str = Field(..., title="Bootstrap")


class AddressCredentials(RootModel[str]):
    root: str = Field(
        ...,
        description=(
            "```\n  ┏━━━━━━━━━━━━┓ ╭───╮ ┏━━━━━━━━━━━━┓\n╾─┫ CREDENTIAL ┣─┤ / ├─┫"
            " CREDENTIAL ┣─╼\n  ┗━━━━━━━━━━━━┛ ╰───╯ ┗━━━━━━━━━━━━┛\n```\n\nOne or two"
            " address credentials, separated by a `/`. The left-side identifies the"
            " payment\npart of the address, and the right-side identifies the"
            " delegation part. Both are optional\nan can be instead a wildcard"
            " (`*`).\n\nExamples:\n-"
            " `addr_vk1x7da0l25j04my8sej5ntrgdn38wmshxhplxdfjskn07ufavsgtkqn5hljl/*`\n-"
            " `*/script1cda3khwqv60360rp5m7akt50m6ttapacs8rqhn5w342z7r35m37`\n-"
            " `dca1e44765b9f80c8b18105e17de90d4a07e4d5a83de533e53fee32e0502d17e/*`\n-"
            " `*/4fc6bb0c93780ad706425d9f7dc1d3c5e3ddbf29ba8486dce904a5fc`\n- `*/*`\n"
        ),
        examples=[
            "addr_vk1x7da0l25j04my8sej5ntrgdn38wmshxhplxdfjskn07ufavsgtkqn5hljl/*",
            "*/script1cda3khwqv60360rp5m7akt50m6ttapacs8rqhn5w342z7r35m37",
            "dca1e44765b9f80c8b18105e17de90d4a07e4d5a83de533e53fee32e0502d17e/*",
            "*/4fc6bb0c93780ad706425d9f7dc1d3c5e3ddbf29ba8486dce904a5fc",
            "*/*",
        ],
        title="Credentials",
    )


class BadRequest(BaseModel):
    model_config = ConfigDict(
        extra="forbid",
    )
    hint: Optional[str] = Field(None, description="Some hint about what went wrong.")


class Datum(BaseModel):
    model_config = ConfigDict(
        extra="forbid",
    )
    datum: str = Field(
        ...,
        description="A serialized Plutus' Data (datum or redeemer).",
        examples=["d87980"],
    )


class DatumType(Enum):
    """
    Categorize the type of datum in the output:

    - `hash`: means that the output only contains a reference to the datum;
    - `inline`: means that the ouput originally contained a full inline datum.

    In both cases however, Kupo returns only a `datum_hash`, and full datums
    can be retrieved via the [`GET /datums/{datum-hash}`](#tag/Datums/paths/~1datums~1{datum-hash}/get)
    endpoint.

    This field is only present when `datum_hash` is not `null`.

    """

    hash = "hash"
    inline = "inline"


class Deleted(BaseModel):
    deleted: conint(ge=0) = Field(
        ..., description="Number of entities effectively deleted."
    )


class Limit(Enum):
    """
    Specify the server behavior when rolling back out of the _safe
    zone_. As mentioned in the user manual, when running Kupo with
    `--prune-utxo` enabled, the server gets rid of spent UTxOs, but it
    only does so after a certain time. That time is exactly `129600`
    slots (or 36h on Mainnet/Testnet). This is because the core
    protocol cannot roll back further than this particular depth and it
    is the point after which it is 100% safe to remove data from the
    database.

    However, this endpoint allows you to break this invariant and
    rollback to points that are even older in the past. As a
    consequence, while syncing, the index may be in a somewhat
    inconsistent state because some inputs spent at a later time may
    not have been recovered during the rollback. This may be surprising
    if you're expecting to see and query those transient inputs after
    rolling back.

    By default, you won't be allowed to rollback beyond the safe zone.
    If, however, you know what you're doing, you're kindly asked to
    pass `unsafe_allow_beyond_safe_zone` as a token of acknowledgment.
    Passing `within_safe_zone` has no effects other than the default.

    Note that, once synchronized again, the index will always be in the
    expected state and problems reflecting reality only occurs _while
    catching up_, after a long rollback.

    """

    unsafe_allow_beyond_safe_zone = "unsafe_allow_beyond_safe_zone"
    within_safe_zone = "within_safe_zone"


class MetadatumInt(BaseModel):
    model_config = ConfigDict(
        extra="forbid",
    )
    value: int = Field(
        ...,
        description=(
            'An integer or arbitrary size.\n\nExample:\n\n```json\n{ "int": 42 }\n```\n'
        ),
        alias="int",
    )


class MetadatumString(BaseModel):
    model_config = ConfigDict(
        extra="forbid",
    )
    string: constr(max_length=64) = Field(
        ...,
        description=(
            "A text string, which is at most 64 bytes.\n\nExample:\n\n```json\n{"
            ' "string": "kupo!" }\n```\n'
        ),
    )


class MetadatumBytes(BaseModel):
    model_config = ConfigDict(
        extra="forbid",
    )
    bytes: constr(max_length=128) = Field(
        ...,
        description=(
            "A base16 byte string, which is at most 64 bytes\n\nExample:\n\n```json\n{"
            ' "bytes": "6b75706f21" }\n```\n'
        ),
    )


class Language(Enum):
    """
    The type of script. `native` refers to pre-Alonzo scripts made of the native DSL to combine keys.
    """

    native = "native"
    plutus_v1 = "plutus:v1"
    plutus_v2 = "plutus:v2"
    plutus_v3 = "plutus:v3"


class Script(BaseModel):
    model_config = ConfigDict(
        extra="forbid",
    )
    language: Language = Field(
        ...,
        description=(
            "The type of script. `native` refers to pre-Alonzo scripts made of the"
            " native DSL to combine keys."
        ),
    )
    script: str = Field(
        ...,
        description="A serialized script (native or Plutus).",
        examples=["4d01000033222220051200120011"],
    )


class Value(BaseModel):
    """
    A (multi-asset) value of a transaction's output.
    """

    model_config = ConfigDict(
        extra="forbid",
    )
    coins: int = Field(..., description="A quantity of Lovelace.", examples=[42])
    assets: Optional[Dict[str, int]] = Field(
        None,
        description="A _key:value_ map of asset identifier → quantity.",
        examples=[
            {
                "1220099e5e430475c219518179efc7e6c8289db028904834025d5b086": 231,
                "289db028904834025d5b085d5b08661220099e5e430475c2195181796.08661220099e": 1,
            }
        ],
    )


class Wildcard(Enum):
    field_ = "*"


class CreatedAt(BaseModel):
    """
    Block reference at which this transaction was included in the ledger.
    """

    model_config = ConfigDict(
        extra="forbid",
    )
    slot_no: conint(ge=0) = Field(
        ..., description="An absolut slot number.", examples=[51540727]
    )
    header_hash: constr(min_length=64, max_length=64) = Field(
        ...,
        description="A blake2b-256 hash digest of a block header.",
        examples=["9d09...31bf"],
    )


class ConnectionStatus(Enum):
    """
    Condition of the connection with the underlying node.
    """

    connected = "connected"
    disconnected = "disconnected"


class Indexes(Enum):
    """
    Behaviour surrounding the database query indexes.
    """

    deferred = "deferred"
    installed = "installed"


class Configuration(BaseModel):
    """
    A summary of hand-picked configuration parameters.
    """

    model_config = ConfigDict(
        extra="forbid",
    )
    indexes: Indexes = Field(
        ..., description="Behaviour surrounding the database query indexes."
    )


class Health(BaseModel):
    """
    An overview of the server & connection status. Note that, when `most_recent_checkpoint` and `most_recent_node_tip` are equal, the index is fully synchronized.
    """

    model_config = ConfigDict(
        extra="forbid",
    )
    connection_status: ConnectionStatus = Field(
        ..., description="Condition of the connection with the underlying node."
    )
    most_recent_checkpoint: Optional[conint(ge=0)] = None
    most_recent_node_tip: Optional[conint(ge=0)] = None
    configuration: Configuration = Field(
        ..., description="A summary of hand-picked configuration parameters."
    )
    version: str = Field(..., description="Current software version.")


class HealthPrometheus(RootModel[str]):
    root: str = Field(
        ...,
        description=(
            "A key-value export compatible with Prometheus.\n\n- `connected` and"
            " `disconnected` (from `connection_status`) are encoded as `1.0` and `0.0`"
            " respectively.\n- `installed` and `deferred` (from"
            " `configuration.indexes`) are encoded as `1.0` and `0.0` respectively.\n-"
            " `version` is not present in the Prometheus metrics.\n"
        ),
        examples={
            "response": {
                "value": (
                    "# TYPE kupo_connection_status gauge\nkupo_connection_status "
                    " 1.0\n\n# TYPE kupo_most_recent_checkpoint"
                    " counter\nkupo_most_recent_checkpoint  294998\n\n# TYPE"
                    " kupo_most_recent_node_tip counter\nkupo_most_recent_node_tip "
                    " 71753381\n\n# TYPE kupo_configuration_indexes"
                    " gauge\nkupo_configuration_indexes  1.0\n"
                )
            }
        },
    )


class RollbackTo(BaseModel):
    """
    A mandatory point to rollback the synchronization to.
    Note that the synchronization will therefore begin starting from the point **immediately after** the provided point!

    > <sup><strong>NOTE (1)</strong></sup> <br/>
    > If you need to query ancestors from any given known point, see [`GET /checkpoints/{slot-no}`](#operation/getCheckpointBySlot)

    > <sup><strong>NOTE (2)</strong></sup> <br/>
    > The `header_hash` is **optional**! However if provided, Kupo will check that it rolls back exactly to the specified point by comparing header hashes.
    > If it's omitted, Kupo will rollback to the given slot number or, any closest ancestor if no point is found at the given slot.

    """

    slot_no: conint(ge=0) = Field(
        ..., description="An absolut slot number.", examples=[51540727]
    )
    header_hash: Optional[constr(min_length=64, max_length=64)] = Field(
        None,
        description="A blake2b-256 hash digest of a block header.",
        examples=["9d09...31bf"],
    )


class ForcedRollback(BaseModel):
    model_config = ConfigDict(
        extra="forbid",
    )
    rollback_to: RollbackTo = Field(
        ...,
        description=(
            "A mandatory point to rollback the synchronization to.\nNote that the"
            " synchronization will therefore begin starting from the point"
            " **immediately after** the provided point!\n\n> <sup><strong>NOTE"
            " (1)</strong></sup> <br/>\n> If you need to query ancestors from any given"
            " known point, see [`GET"
            " /checkpoints/{slot-no}`](#operation/getCheckpointBySlot)\n\n>"
            " <sup><strong>NOTE (2)</strong></sup> <br/>\n> The `header_hash` is"
            " **optional**! However if provided, Kupo will check that it rolls back"
            " exactly to the specified point by comparing header hashes.\n> If it's"
            " omitted, Kupo will rollback to the given slot number or, any closest"
            " ancestor if no point is found at the given slot.\n"
        ),
    )
    limit: Optional[Limit] = Field(
        "within_safe_zone",
        description=(
            "Specify the server behavior when rolling back out of the _safe\nzone_. As"
            " mentioned in the user manual, when running Kupo with\n`--prune-utxo`"
            " enabled, the server gets rid of spent UTxOs, but it\nonly does so after a"
            " certain time. That time is exactly `129600`\nslots (or 36h on"
            " Mainnet/Testnet). This is because the core\nprotocol cannot roll back"
            " further than this particular depth and it\nis the point after which it is"
            " 100% safe to remove data from the\ndatabase.\n\nHowever, this endpoint"
            " allows you to break this invariant and\nrollback to points that are even"
            " older in the past. As a\nconsequence, while syncing, the index may be in"
            " a somewhat\ninconsistent state because some inputs spent at a later time"
            " may\nnot have been recovered during the rollback. This may be"
            " surprising\nif you're expecting to see and query those transient inputs"
            " after\nrolling back.\n\nBy default, you won't be allowed to rollback"
            " beyond the safe zone.\nIf, however, you know what you're doing, you're"
            " kindly asked to\npass `unsafe_allow_beyond_safe_zone` as a token of"
            " acknowledgment.\nPassing `within_safe_zone` has no effects other than the"
            " default.\n\nNote that, once synchronized again, the index will always be"
            " in the\nexpected state and problems reflecting reality only occurs"
            " _while\ncatching up_, after a long rollback.\n"
        ),
    )


class Pattern(
    RootModel[
        Union[
            Wildcard,
            str,
            constr(pattern=r"[0-9a-f]{56}\.(.*|[0-9a-f]{2,64})"),
            constr(pattern=r"(.*|[0-9]+)@[0-9a-f]{64}"),
        ]
    ]
):
    root: Union[
        Wildcard,
        str,
        constr(pattern=r"[0-9a-f]{56}\.(.*|[0-9a-f]{2,64})"),
        constr(pattern=r"(.*|[0-9]+)@[0-9a-f]{64}"),
    ] = Field(
        ...,
        description=(
            "A matching pattern for addresses, assets or transactions.\n\nSee"
            " [Patterns](#section/Patterns) for more details.\n"
        ),
    )


class Point(BaseModel):
    model_config = ConfigDict(
        extra="forbid",
    )
    slot_no: conint(ge=0) = Field(
        ..., description="An absolut slot number.", examples=[51540727]
    )
    header_hash: constr(min_length=64, max_length=64) = Field(
        ...,
        description="A blake2b-256 hash digest of a block header.",
        examples=["9d09...31bf"],
    )


class SpentAt(BaseModel):
    model_config = ConfigDict(
        extra="forbid",
    )
    slot_no: conint(ge=0) = Field(
        ..., description="An absolut slot number.", examples=[51540727]
    )
    header_hash: constr(min_length=64, max_length=64) = Field(
        ...,
        description="A blake2b-256 hash digest of a block header.",
        examples=["9d09...31bf"],
    )
    transaction_id: Optional[constr(min_length=64, max_length=64)] = None
    input_index: Optional[conint(ge=0)] = None
    redeemer: Optional[str] = None


class Match(BaseModel):
    model_config = ConfigDict(
        extra="forbid",
    )
    transaction_index: conint(ge=0) = Field(
        ...,
        description="The index of the transaction within the block including it.",
        examples=[14],
    )
    transaction_id: constr(min_length=64, max_length=64) = Field(
        ...,
        description="A blake2b-256 hash digest of a transaction body.",
        examples=["35d8...4e09"],
    )
    output_index: conint(ge=0) = Field(
        ...,
        description="The index of the output within the transaction carrying it.",
        examples=[2],
    )
    address: str = Field(..., description="A Cardano address, in any era.")
    value: Value
    datum_hash: Optional[constr(min_length=64, max_length=64)] = Field(
        ..., description="A blake2b-256 hash digest of a Plutus' datum, if any."
    )
    datum: Optional[str] = Field(
        None,
        description=(
            "The resolved datum, if available. The field is only and always present"
            " (yet may be `null`) if `?resolve_hashes` was set."
        ),
    )
    datum_type: Optional[DatumType] = None
    script_hash: Optional[constr(min_length=56, max_length=56)] = Field(
        ...,
        description="A blake2b-224 hash digest of a Native or Plutus script, if any.",
    )
    script: Optional[Script] = Field(
        None,
        description=(
            "The resolved script, if available. The field is only and always present"
            " (yet may be `null`) if `?resolve_hashes` was set."
        ),
    )
    created_at: CreatedAt = Field(
        ...,
        description=(
            "Block reference at which this transaction was included in the ledger."
        ),
        title="Point",
    )
    spent_at: Optional[SpentAt] = Field(
        ...,
        description=(
            "Block reference at which this transaction input was spent, if any."
        ),
    )


class Metadata(BaseModel):
    model_config = ConfigDict(
        extra="forbid",
    )
    hash: constr(min_length=64, max_length=64) = Field(
        ...,
        description="A blake2b-256 hash digest of the raw serialized data",
        examples=["cd6a5d31bf9d309706b92ad83402e682fdab9fc889b1b63565ee3de14e09dedf"],
    )
    raw: str = Field(
        ...,
        description=(
            "A [CBOR](https://www.rfc-editor.org/rfc/rfc8949.html)-encoded binary"
            " payload."
        ),
    )
    schema_: Dict[
        str,
        Union[
            MetadatumInt, MetadatumString, MetadatumBytes, MetadatumList, MetadatumMap
        ],
    ] = Field(
        ...,
        alias="schema",
        description=(
            "A high-level description of the raw data. The top-level object is an"
            " object where all keys are integers (possibly negative) and points to"
            " objects representing one of 5 primitives:\n\n- int\n- string\n- bytes\n-"
            " list\n- map\n\nThis schema is meant to give a faithful representation of"
            " the underlying [CBOR](https://www.rfc-editor.org/rfc/rfc8949.html)"
            " encoding, which is slightly more expressive than pure JSON (e.g. keys of"
            " maps can be arbitrary CBOR objects).\n\nThis is why it is generally not"
            " possible to ensure a 1:1 conversion between low-level CBOR and JSON, and"
            " why this intermediate representation is necessary to safely represent"
            " **any** metadata object.\n"
        ),
        examples=[
            {
                "64": {"string": "some text"},
                "32": {"int": 42},
                "16": {
                    "map": [
                        {
                            "k": {"string": "numbers"},
                            "v": {
                                "list": [{"int": 1}, {"int": 2}, {"int": 4}, {"int": 8}]
                            },
                        },
                        {
                            "k": {"string": "alphabet"},
                            "v": {
                                "map": [
                                    {"k": {"string": "A"}, "v": {"int": 65}},
                                    {"k": {"string": "B"}, "v": {"int": 66}},
                                    {"k": {"string": "C"}, "v": {"int": 67}},
                                ]
                            },
                        },
                    ]
                },
                "-8": {"bytes": "48656c6c6f2c2043617264616e6f21"},
            }
        ],
    )


class MetadatumList(BaseModel):
    model_config = ConfigDict(
        extra="forbid",
    )
    list: List[
        Union[
            MetadatumInt, MetadatumString, MetadatumBytes, MetadatumList, MetadatumMap
        ]
    ] = Field(
        ...,
        description=(
            "A (possibly heterogeneous) list of metadatum\n\nExample:\n\n```json\n{\n "
            ' "list": [\n    {\n      "int": 14\n    },\n    {\n      "list": [\n      '
            '  { "string": "kupo!" },\n        { "bytes": "6b75706f21" }\n      ]\n   '
            " }\n  ]\n}\n```\n"
        ),
    )


class MapItem(BaseModel):
    k: Union[MetadatumInt, MetadatumString, MetadatumBytes, MetadatumList, MetadatumMap]
    v: Union[MetadatumInt, MetadatumString, MetadatumBytes, MetadatumList, MetadatumMap]


class MetadatumMap(BaseModel):
    model_config = ConfigDict(
        extra="forbid",
    )
    map: List[MapItem] = Field(
        ...,
        description=(
            "A list of key:value objects. Both keys and values can be any sort"
            ' metadatum.\n\nExample:\n\n```json\n{\n  "map": [\n    {\n      "k": {'
            ' "int": 14 },\n      "v": {\n        "list": [\n          { "string":'
            ' "kupo!" },\n          { "bytes": "6b75706f21" }\n        ]\n      }\n   '
            " }\n  ]\n}\n```\n"
        ),
    )


Metadata.model_rebuild()
